#include <iostream>
#include <string>

using namespace std;

// Abstraktni trida - kazda trida, ktera ma abstraktni metodu
// je abstraktni. Nelze vytvorit instanci abstraktni tridy (viz
// funkce main.
class GObject{

public:
	virtual void printObsah(){
		cout << "Zadny obsah" << endl;
	}
	// Deklarace abstraktni metody - musi byt virtual a " = 0"
	virtual void printObvod() = 0;
	void printInfo(){
		cout << "GObject" << endl;
	}
};

// Trida GKruh dedi ze zakladni abstraktni tridy
class GCircle : public GObject {
	double r;
public:
	GCircle(double _r): r(_r) {};
	// virtualni metoda, lze ji volat "pozdni vazbou"
	virtual void printObsah(){
		cout << "Obsah kruhu: " << (3.14 * r *r) << endl;
	}
	// virtualni metoda, lze ji volat "pozdni vazbou"
	virtual void printObvod(){
		cout << "Ovod kruhu: " << (2 * 3.14 * r) << endl;
	}
	void printInfo(){
		cout << "GCircle" << endl;
	}
};

// Dalsi objekt dedici z abstraktni tridy
class GSquare : public GObject {
	double a;
public:
	GSquare(double _a): a(_a){};
	// Metoda neni virtualni, nelze ji volat pozdni vazbou
	void printObsah(){
		cout << "Obsah ctverce: " << a * a << endl;
	}
	// Nasledujici metoda schvalne nema delat vubec nic, ale
	// protoze je virtualni a jeji predek je abstraktni, tak
	// musi mit telo, i kdyz je treba prazdne!
	virtual void printObvod(){
		// Metoda je prazdna. Jeji funkci nepotrebujeme,
		// ale v predkovi je abstratni, takze byt definova
		// musi!!!
	}
};

// Vytvorim tridu pro reprezentaci vyjimky
class Vyjimka {
	string text;
public:
	Vyjimka(string text){
		this->text = text;
	}
	string &getError(){
		return text;
	}
};

// V pripade, ze dojde k pokusu o deleni nulou, dojde k vyjimce
double vydel(double a, double b) throw (Vyjimka){
	if(b == 0){
		int *cislo = new int(4); // !!!
		Vyjimka v(string("Deleni nulou!!!"));
		throw v; // Pozor v tomto miste dojde k opusteni aktualniho
		// bloku kodu. Nasledujici radek se jiz neprovede, takze nedojde
		// k uvolneni pameti a ziskame tak memory leak!!!
		// Ve vyjimce se vraci kopie objektu "v".
		delete cislo; // !!!
	}
	return a/b;
}

// Seznam vyjimek, ktere mohou ve funkci vzniknout neni povinny
double vydel2(double a, double b) /*throw (Vyjimka *)*/{
	if(b == 0){
		// Ve vyjimce se vraci ukazatel na novy objekt typu Vyjimka.
		// Po odchyceni vyjimky, se musi Vyjimka dealokovat!!!
		throw new Vyjimka("Deleni nulou!!!");
	}
	return a/b;
}

// Ovsem pokud dojde k vyhozeni vyjimky, ktera v seznamu neni uvedena,
// je to chyba, ktera se da detekovat a lze na ni vytvorit vlastni reakci
// - viz fukce moje_unexpected()
double vydel3(double a, double b) throw (Vyjimka){
	if(b == 0){
		throw 3;
	}
	return a/b;
}

// Funkce, ktera se ve funkci main preda jako ukzatel standardni funkci
// set_unexpected a lze tak vytvorit vlastni reakci na neocekavanou
// vyjimku.
void moje_unexpected(){
	cout << "Neocekavana vyjimka" << endl;
}

// Stejne jako v predchozim pripade, lze vytvorit vlastni reakci na
// neodchycenou vyjimku.
void moje_terminate(){
	cout << "Neodchycena vyjimka" << endl;
}

int main(){
	//GObject o; // Abstraktni trida, nelze vytvorit jeji instanci!!
	GSquare s(3); // Tady je vse v poradku
	GCircle k(3);

	// Obe nove instance trid GSquare a GCircle lze diky spolecnemu predkovi
	// ulozit do stejneho pole. Zde se jedna o pole ukazatelu.
	GObject *pole[2];

	// Do polozek pole vlozim ukazatele na obe instance objektu
	pole[0] = &s;
	pole[1] = &k;

	cout << endl << "Ukazka prace s pozdni vazbou:" << endl << endl;
	// Zajimave je nasledujici chovani:
	// Zavolam metody printObvod() a printObsah() pomoci pozdni vazby.
	// To znamena, ze mam ukazatele na predka, ktery ukazuje na potomka.
	// Na tomto ukazateli zavolam metodu, ktera je implementovana jak
	// v predkovi, tak v potomkovi. Pozdni vazba zpusobi, ze se zavola
	// metoda na predkovi:
	pole[1]->printObvod();
	pole[1]->printObsah();
	pole[0]->printObsah();
	// Pokud nepouziji pozdni vazbu, zavola se metoda na predkovi:
	pole[0]->printInfo();
	// Aby to bylo kompletni zavolame metodu printObvod - samozrejme se
	// zavola na potomkovi GSquare, ale je prazdna, takze se nedeje nic.
	pole[0]->printObvod();

	cout << endl << "-------------------------------------------------------" << endl << endl;

	cout << "Ukazka prace s vyjimkami" << endl << endl;
	try{
		vydel(5, 0);
	} catch (Vyjimka v){
		cout<< v.getError() << endl;
	}

	try{
		vydel2(5, 0);
	} catch (Vyjimka *v){
		cout<< v->getError() << endl;
		// Obdrzel jsem ukazatel na objekt, ktery uz dale nebudu pouzivat,
		// protoze byl dynamicky alokovan, musim jej dealokovat:
		delete v;
	}

	// Odchyceni jakekoliv vyjimky:
	try{
		vydel2(5, 0);
	} catch(...){
		cout << endl << "Dorazila nejaka vyjimka..." << endl << endl;
	}
	// Nastavim svou funkci
	set_unexpected(moje_unexpected);
	try{
		vydel3(5, 0);
	} catch (Vyjimka *v){
		cout<< v->getError() << endl;
		delete v;
	}

	set_terminate(moje_terminate);
	vydel(5, 0);
	return 0;
}
